/*--------------------------------------------------------------------
 * $Id: cloud.c,v 1.26 2002/12/22 00:55:04 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ckdfu.h>
#include "uvspec.h"
#include "cloud.h"
#include "../src_c/ascii.h"
#include "../src_c/numeric.h"
#include "../src_c/miecalc.h"
#include "../src_f/f77-uscore.h"

#if HAVE_YANG56
#include "../src_c/yang56.h"
#endif

#define EPSILON 1E-6

/************************************/
/* prototypes of internal functions */
/************************************/

static int interpolate_legendre (float **legen, int *nleg, int nreff,
				 double r0, double dr,
				 double *extinc, double *albedo,
				 double reff,
				 float **coeffc, int *ncoeffc, 
				 float *ext, float *alb);

static int calloc_cld_microphysics (cld_out_struct *out, int nlambda, int nlev);

static int ic_yang (float wavelength, int nlyr, int habit,
		    char *path,
		    float *iwc, float *reff, 
		    float *dtau, 
		    float *f, float *g1, float *g2, float *ssa, 
		    float *zd, int layer);


/**************************************************************/
/* Setup water clouds.                                        */
/**************************************************************/

int setup_wcloud (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0;

  /* if water cloud microphysical properties profile specified */
  if (strlen(input.filename[FN_WCLOUD]) > 0) {

    /* Read microphysical properties from input.filename[FN_WCLOUD] */
    /* and convert to optical properties.                           */

    if (input.verbose) {
      fprintf (stderr, " ... reading water cloud properties from %s\n", 
	       input.filename[FN_WCLOUD]);
    }

    /* Read water cloud microphysical properties and apply */
    /* default conversion to optical properties            */
    status = read_and_convert_wc_file (input, output, 
				       input.filename[FN_WCLOUD], 
				       &(output->wc));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_and_convert_wc_file (%s)\n", 
	       status, input.filename[FN_WCLOUD]);
      return status;
    }
  }
  else {   
    
    /* Read file holding filenames of layer cloud optical properties */    

    if (input.wc.properties == WCPROP_EXPLICIT) {

      /* just to make sure, check if filename has been defined */
      if (strlen(input.filename[FN_WCLOUD_FILES]) == 0) {
	fprintf (stderr, "Error, inconsistent cloud properties. This is not\n");
	fprintf (stderr, "supposed to happen and indicates a coding error.\n");
	return -1;
      }

      if (input.atm.altitude != 0) {
	fprintf (stderr, "Sorry, option 'altitude' currently not implemented in combination\n");
	fprintf (stderr, "with 'wc_files'\n");
	return -1;
      }
      
      /* determine number of levels; required for memory allocation */
      if ((status = ASCII_checkfile (input.filename[FN_WCLOUD_FILES], 
				     &rows, &min_columns, &max_columns, &max_length)) != 0) {
	fprintf (stderr, "Error %d reading %s\n", status, input.filename[FN_WCLOUD_FILES]);
	return status;
      }

      /* allocate memory */
      status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, rows);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out struct\n", status);
	return status;
      }

      /* read optical properties profile, */
      /* allocate memory for moments      */

      status = read_optprop_files (input.filename[FN_WCLOUD_FILES],
				   output->wl.lambda_t, output->wl.nlambda_t, 
				   output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				   output->wc.zd, output->wc.nlev,
				   output->wc.optprop.dtau, 
				   output->wc.optprop.ssa, 
				   output->wc.optprop.g1,
				   output->wc.optprop.moment, output->wc.optprop.nmom);
      if (status != 0) {
	fprintf (stderr, "Error %d reading cloud optical property files: %s\n", 
		 status, input.filename[FN_WCLOUD_FILES]);
	return status;
      }
      
      
      /* increase number of moments if more found */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->wc.nlev-1; lc++)
	  if (output->wc.optprop.nmom[iv][lc]-1 > output->atm.nmom )
	    output->atm.nmom = output->wc.optprop.nmom[iv][lc]-1;
    }
    else {    /* if neither of the cloud property files defined */

      /* allocate memory, even if no cloud is defined; */
      /* it may be required by other functions;        */

      status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, 2);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out-struct\n", status);
	return status;
      }

      output->wc.zd[0] = output->atm.zd[0];
      output->wc.zd[1] = output->atm.zd[output->atm.nlev-1];
    }
  }

  /* overwrite these properties with user-defined optical thickness, ssa, etc */
  status = apply_user_defined_properties_to_cld (input.wc, output->atm, output->wl, 
						 input.atm.altitude, &(output->wc));
  if (status!=0) {
    fprintf (stderr, "Error %d overwriting cloud properties with user-defined values\n",
	     status);
    return status;
  }



  /* add the water cloud levels to the common profile */
  set_common_z_grid (output->wc.zd, output->wc.nlev, 
		     &output->atm.zd_common, &output->atm.nlev_common);

  return 0;
}




/**************************************************************/
/* Setup ice clouds.                                          */
/**************************************************************/

int setup_icloud (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0;

  /* if ice cloud microphysical properties profile specified */
  if (strlen(input.filename[FN_ICLOUD]) > 0) {

    /* Read microphysical properties from input.filename[FN_ICLOUD] */
    /* and convert to optical properties.                           */

    if (input.verbose)
      fprintf (stderr, " ... reading ice cloud microphysical properties from %s\n", 
	       input.filename[FN_ICLOUD]);
    
    /* Read ice cloud microphysical properties */
    /* convert to optical properties           */
    status = read_and_convert_ic_file (input, output, 
				       input.filename[FN_ICLOUD], 
				       &(output->ic));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_and_convert_ic_file (%s)\n", 
	       status, input.filename[FN_ICLOUD]);
      return status;
    }
  }
  else {

    if (input.ic.properties == ICPROP_EXPLICIT) {
      /* Read file holding filenames of layer cloud optical properties */
      
      /* just to make sure, check if filename has been defined */
      if (strlen(input.filename[FN_ICLOUD_FILES]) == 0) {
	fprintf (stderr, "Error, inconsistent cloud properties. This is not\n");
	fprintf (stderr, "supposed to happen and indicates a coding error.\n");
	  return -1;
      }

      if (input.atm.altitude != 0) {
	fprintf (stderr, "Sorry, option 'altitude' currently not implemented in combination\n");
	fprintf (stderr, "with 'ic_files'\n");
	return -1;
      }
      

      /* determine number of levels; required for memory allocation */
      if ((status = ASCII_checkfile (input.filename[FN_ICLOUD_FILES], 
				     &rows, &min_columns, &max_columns, &max_length)) != 0) {
	fprintf (stderr, "Error %d reading %s\n", status, input.filename[FN_ICLOUD_FILES]);
	return status;
      }

      /* allocate memory */
      status = calloc_cld_out (&(output->ic), output->wl.nlambda_t, rows);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out struct\n", status);
	return status;
      }
      
      /* read optical properties profile, */
      /* allocate memory for moments      */

      status = read_optprop_files (input.filename[FN_ICLOUD_FILES],
				   output->wl.lambda_t, output->wl.nlambda_t,
				   output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				   output->ic.zd, output->ic.nlev,
				   output->ic.optprop.dtau, 
				   output->ic.optprop.ssa, 
				   output->ic.optprop.g1,
				   output->ic.optprop.moment, output->ic.optprop.nmom);
      if (status != 0) {
	fprintf (stderr, "Error %d reading cloud optical property files: %s\n", 
		 status, input.filename[FN_ICLOUD_FILES]);
	return status;
      }
    
      /* increase number of moments if more found */
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	for (lc=0; lc<output->ic.nlev-1; lc++)
	  if (output->ic.optprop.nmom[iv][lc]-1 > output->atm.nmom )
	    output->atm.nmom = output->ic.optprop.nmom[iv][lc]-1;
    }
    else { /* if neither of the cloud property files defined */
      
      /* allocate memory, even if no cloud is defined; */
      /* it may be required by other functions.        */

      status = calloc_cld_out (&(output->ic), output->wl.nlambda_t, 2);
      if (status!=0) {
	fprintf (stderr, "Error %d allocating memory for cld_out-struct\n", status);
	return status;
      }

      output->ic.zd[0] = output->atm.zd[0];
      output->ic.zd[1] = output->atm.zd[output->atm.nlev-1];
    }
  }


  /* overwrite these properties with user-defined optical thickness, ssa, etc */
  status = apply_user_defined_properties_to_cld (input.ic, output->atm, output->wl, 
						 input.atm.altitude, &(output->ic));
  if (status!=0) {
    fprintf (stderr, "Error %d overwriting cloud properties with user-defined values\n",
	     status);
    return status;
  }

  /* add the ice cloud levels to the common profile */
  set_common_z_grid (output->ic.zd, output->ic.nlev, 
		     &output->atm.zd_common, &output->atm.nlev_common);

  return 0;
}



/*********************************/
/* compare two altitude grids    */
/*********************************/

int compare_levels (float *z1, int n1,
		    float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "Error, number of layers differing\n");
    fprintf (stderr, "%d vs %d\n", n1, n2);

    return -1;
  }

  for (kc=0; kc<n1; kc++)
    if (fabs((float) (z1[kc]) - z2[kc]) > EPSILON * z1[kc]) {
      fprintf (stderr, "Error, difference at level %d, altitude %g vs. %g\n",	
	       kc, z1[kc], z2[kc]);
      return -1;
    }

  return 0;  /* if o.k. */
}


static int calloc_cld_microphys_struct (cld_microphys_struct *out, int nlev) 
{
  out->lwc  = calloc (nlev, sizeof(float));
  out->effr = calloc (nlev, sizeof(float));
  out->alloc = 1;

  return 0;
}


int calloc_optprop (optprop_struct *out, int nlambda, int nlev) 
{
  int iv=0, lc=0;

  out->nlam = nlambda;
  
  out->dtau = calloc (nlambda, sizeof (float *));
  out->ssa  = calloc (nlambda, sizeof (float *));
  out->fdel = calloc (nlambda, sizeof (float *));

  /* parameters for double HG */
  out->g1   = calloc (nlambda, sizeof (float *));
  out->g2   = calloc (nlambda, sizeof (float *));
  out->ff   = calloc (nlambda, sizeof (float *));
  
  
  for (iv=0; iv<nlambda; iv++) {
    out->dtau[iv] = calloc (nlev, sizeof(float));
    out->ssa [iv] = calloc (nlev, sizeof(float));
    out->fdel[iv] = calloc (nlev, sizeof(float));
    out->g1  [iv] = calloc (nlev, sizeof(float));
    out->g2  [iv] = calloc (nlev, sizeof(float));
    out->ff  [iv] = calloc (nlev, sizeof(float));

    for (lc=0; lc<nlev; lc++)
      out->ff[iv][lc] = 1.0;
  }
  
  out->nmom   = calloc (nlambda, sizeof (int *));
  out->moment = calloc (nlambda, sizeof (float **)); 
  for (iv=0; iv<nlambda; iv++) {
    out->nmom[iv]   = calloc (nlev, sizeof (int ));
    out->moment[iv] = calloc (nlev, sizeof (float *));
    
    /* initialize nmom to -1; nmom==0 means an isotropic phase function */
    for (lc=0; lc<nlev; lc++)
      out->nmom[iv][lc] = 0;
  }
  
  out->alloc = 1;

  return 0;
}



int free_optprop (optprop_struct *optprop)
{
  int iv=0;

  for (iv=0; iv<optprop->nlam; iv++) {
    free (optprop->g1[iv]);
    free (optprop->g2[iv]);
    free (optprop->ff[iv]);
    free (optprop->ssa[iv]);
    free (optprop->dtau[iv]);
    free (optprop->nmom[iv]);
    free (optprop->moment[iv]);
  }

  free (optprop->g1);
  free (optprop->g2);
  free (optprop->ff);
  free (optprop->ssa);
  free (optprop->dtau);
  free (optprop->nmom);
  free (optprop->moment);

  return 0;
}




static int calloc_cld_ssprop_struct (cld_ssprop_struct *out, int nlambda) 
{

  out->extinc = calloc (nlambda, sizeof(double *));
  out->albedo = calloc (nlambda, sizeof(double *));
  out->legen  = calloc (nlambda, sizeof(float **));
  out->nleg   = calloc (nlambda, sizeof(int *));
  out->r0     = calloc (nlambda, sizeof(double));
  out->dr     = calloc (nlambda, sizeof(double));
  out->nreff  = calloc (nlambda, sizeof(int));
  
  out->alloc  = 1;

  return 0;
}



/***************************************/
/* allocate memory for a cld_out_struct */
/***************************************/

int calloc_cld_out (cld_out_struct *out, int nlambda, int nlev) 
{
  int status=0;

  out->zd     = calloc (nlev, sizeof(float));
  out->nlev   = nlev;
  out->newsiz = 1;


  status = calloc_cld_microphys_struct (&(out->microphys), out->nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud microphysics\n", status);
    return status;
  }
  
  status = calloc_optprop (&(out->optprop), nlambda, out->nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud optical properties\n", status);
    return status;
  }

  status = calloc_cld_ssprop_struct (&(out->ssprop), nlambda);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud single scattering properties\n", 
	     status);
    return status;
  }

  return 0;
}



/***********************************************************/
/* allocate memory for a cld_out_struct, microphysics only */
/***********************************************************/

static int calloc_cld_microphysics (cld_out_struct *out, int nlambda, int nlev) 
{
  int status=0;

  out->zd     = calloc (nlev, sizeof(float));
  out->nlev   = nlev;
  out->newsiz = 1;
  
  status = calloc_cld_microphys_struct (&(out->microphys), out->nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cloud microphysics\n", status);
    return status;
  }

  return 0;
}




int cp_optprop (optprop_struct *target, optprop_struct *source, int nlev)
{
  int iv=0, lev=0;

  if (target->alloc && source->alloc) {
    
    /* check if number of wavelength bands equal */
    if (target->nlam != source->nlam) {
      fprintf (stderr, "Error: cp_optprop(), nlam differs between source and target\n");
      fprintf (stderr, "       source: %d,  target: %d\n", 
	       source->nlam, target->nlam);
      return -1;
    }
    
    for (lev=0; lev<nlev-1; lev++) {
      for (iv=0; iv<source->nlam; iv++) {
	target->dtau [iv][lev] = source->dtau [iv][lev];
	target->ssa  [iv][lev] = source->ssa  [iv][lev];
	
	target->g1   [iv][lev] = source->g1   [iv][lev];
	target->g2   [iv][lev] = source->g2   [iv][lev];
	target->ff   [iv][lev] = source->ff   [iv][lev];
      }
    }
  }
  else {
    fprintf (stderr, "Warning, not copying optical properties because\n");
    fprintf (stderr, "memory for source or target not allocated\n");
  }

  return 0;
}
    

/************************************************/
/* copy contents of a cld_out_struct to another; */
/* NO memory allocation is done here!!!         */
/************************************************/

int cp_cld_out (cld_out_struct *target, cld_out_struct source, int copy_ssprop)
{
  int iv=0, ir=0, k=0, lev=0, status=0;

  if (target->nlev != source.nlev) {
    fprintf (stderr, "Error: cp_cld_out(), nlev differs between source and target\n");
    fprintf (stderr, "       source: %d,  target: %d\n", source.nlev, target->nlev);
    return -1;
  }

  target->newsiz = source.newsiz;

  /* copy microphysical properties */
  for (lev=0; lev<source.nlev; lev++) {
    target->zd[lev] = source.zd[lev];

    if (target->microphys.alloc) {
      target->microphys.lwc  [lev] = source.microphys.lwc  [lev];
      target->microphys.effr [lev] = source.microphys.effr [lev];
    }
  }

  /* copy single scattering properties */
  if (source.ssprop.alloc && copy_ssprop)  {
    target->ssprop.alloc = source.ssprop.alloc;

    for (iv=0; iv<source.optprop.nlam; iv++) {
      target->ssprop.nreff[iv] = source.ssprop.nreff[iv];
      target->ssprop.r0[iv]    = source.ssprop.r0[iv];
      target->ssprop.dr[iv]    = source.ssprop.dr[iv];

      for (ir=0; ir<source.ssprop.nreff[iv]; ir++) {
	target->ssprop.extinc [iv][ir] = source.ssprop.extinc [iv][ir];
	target->ssprop.albedo [iv][ir] = source.ssprop.albedo [iv][ir];

	target->ssprop.nleg [iv][ir]   = source.ssprop.nleg [iv][ir];
	
	for (k=0; k<source.ssprop.nleg[iv][ir]; k++)
	  target->ssprop.legen [iv][ir][k] = source.ssprop.legen [iv][ir][k];
      }
    }
  }
  

  /* copy optical properties */
  status = cp_optprop (&(target->optprop), &(source.optprop), source.nlev);
  if (status!=0) {
    fprintf (stderr, "Error %d copying cloud optical properties\n", status);
    return status;
  }

  
  return 0;
}




/* Interpolate the Legendre coefficients which are given on a grid of  */
/* effective radii r0, r0 + dr, ..., r0 + (nreff-1) * dr to a given    */
/* radius reff; memory for the array of Legendre coefficients coeffc[] */
/* is allocated automatically.                                         */

static int interpolate_legendre (float **legen, int *nleg, int nreff,
				 double r0, double dr,
				 double *extinc, double *albedo,
				 double reff,
				 float **coeffc, int *ncoeffc, float *ext, float *alb)
{
  int i=0, k=0;
  int lower=0, upper=0;
  int min=0, max=0;

  double reffl=0;

  if (reff<r0 || reff> r0 + (double) (nreff-1) * dr) {
    fprintf (stderr, "Error, reff = %f out of range (%f - %f)\n",
	     reff, r0, r0 + (double) (nreff-1) * dr);
    return -1;
  }

  lower = (int) ((reff - r0) / dr);
  upper = lower+1;
  
  reffl = r0 + (double) lower * dr;

  if (lower==nreff-1) {   /* upper boundary value */
    *coeffc  = calloc (nleg[nreff-1], sizeof(float));
    *ncoeffc = nleg[nreff-1];

    for (k=0; k<*ncoeffc; k++)
      (*coeffc)[k] = legen[nreff-1][k];
    
    *alb = albedo[lower];
    *ext = extinc[lower];

    return 0;
  }

  /* otherwise interpolation required */
  min = (nleg[lower] < nleg[upper] ? nleg[lower] : nleg[upper]);
  max = (nleg[lower] > nleg[upper] ? nleg[lower] : nleg[upper]);
  
  *ncoeffc = max;
  *coeffc  = calloc (*ncoeffc, sizeof(float));

  for (i=0; i<min; i++)
    (*coeffc)[i] = legen[lower][i] + (reff - reffl) / dr * (legen[upper][i] - legen[lower][i]);
    
  for (i=min; i<*ncoeffc; i++)
    if (nleg[lower] < nleg[upper])
      (*coeffc)[i] = (reff - reffl) / dr * (legen[upper][i]);
    else
      (*coeffc)[i] = legen[lower][i] - (reff - reffl) / dr * legen[lower][i];

  
  /* extinction and single scattering albedo */
  *ext = extinc[lower] + (reff - reffl) / dr * (extinc[upper] - extinc[lower]);
  *alb = albedo[lower] + (reff - reffl) / dr * (albedo[upper] - albedo[lower]);

  return 0;
}


/*********************************************************************/
/* Read cloud data from file; microphysical properties are expected  */
/* in three columns:                                                 */
/*   column WCZD:    z [km]                                          */
/*   column WCLWC:   lwc [g / m3]                                    */
/*   column WCEFFR:  effective radius [micron]                       */
/* Memory for microphysical and optical properties is allocated      */
/* automatically. This function can be used for ice and water clouds */
/* because all output is written to cld_out_struct *cld which can    */
/* either be output->wc or output->ic                                */
/*********************************************************************/

int read_cld_file (input_struct input, int nlambda,
		   char *filename, cld_out_struct *cld,
		   int optical) 
{
  int lev=0, rows=0, min_columns=0, max_columns=0, status=0;
  float **cld_data=NULL;
  
  
  status = ASCII_file2float (filename,
			     &rows, &max_columns, &min_columns, 
			     &cld_data);
  if (status!=0) {
    fprintf (stderr, "Error %d opening file %s\n", status, filename);
    return status;
  }
  
  if (max_columns!=min_columns) {
    fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
    fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
  }  

  if (min_columns<3) {
    fprintf (stderr, "Error, found less than 3 columns in %s\n",
	     filename);
    return -1;
  }
  

  /* allocate memory for 1D cloud structure */

  if (optical) {
    status = calloc_cld_out (cld, nlambda, rows);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
      return status;
    }
  }
  else {
    status = calloc_cld_microphysics (cld, nlambda, rows);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
      return status;
    }
  }

  for (lev=0; lev<cld->nlev; lev++) {
    cld->zd            [lev] = cld_data[lev][CLD_ZD];
    cld->microphys.lwc [lev] = cld_data[lev][CLD_LWC];
    cld->microphys.effr[lev] = cld_data[lev][CLD_EFFR];
  }
  
  
  /* free temporary memory */
  ASCII_free_float(cld_data, rows);
  
  /* newsiz is a flag, telling the Fortran function wcloud() if a */
  /* new liquid water content and/or effective radius are given   */
  /* (newsiz=1) or if the values haven't changed since the last   */
  /* call (newsiz=0).                                             */ 
  cld->newsiz = 1;
  
  
  return 0;
}
  




/* Read water cloud data from file using read_cld_file() and convert */
/* microphysical properties to optical properties                    */

int read_and_convert_wc_file (input_struct input, output_struct *output,
			      char *filename, cld_out_struct *wc)
{
  int iv=0, nlyr=0, status=0;
  int nstring = (int) strlen(input.filename[FN_PATH]);

  void  F77_FCN (wcloud) (float* lambda_t, int* newsiz, int* nlyr,
			  char* filepath, int* nstring, float* lwc, float* wceffr, 
			  float* tmp_wc_dtau, float* tmp_wc_gg, float* tmp_wc_ssa, float* zd,
			  int *wclyr);

  

  status = read_cld_file (input, output->wl.nlambda_t, filename, wc, 1);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, input.filename[FN_WCLOUD]);
    return status;
  }
  

  switch (input.wc.properties) {
  case WCPROP_HU:
    
    switch (input.ck_scheme) {
    case CK_FU:
      
      /* call Fu and Liou code by Fred Rose */
      
#if HAVE_FULIOU
      status = ckdfucld (wc->zd, 
			 wc->microphys.effr, 
			 wc->microphys.lwc, 
			 wc->nlev,
			 wc->optprop.dtau, 
			 wc->optprop.g1, 
			 wc->optprop.ssa);
#else
      fprintf (stderr, "Error, Fu and Liou not supported!\n");
      return -1;
#endif
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ckdfucld()\n", status);
	return status;
      }
      
      break;
      
    default:   /* old uvspec wcloud.f for Hu and Stamnes */
      
      nlyr = output->wc.nlev-1;

      for (iv=0; iv<output->wl.nlambda_t; iv++)
	F77_FCN (wcloud) (&output->wl.lambda_t[iv], &(wc->newsiz), &nlyr,
			  input.filename[FN_PATH], &nstring, 
			  wc->microphys.lwc, wc->microphys.effr, 
			  wc->optprop.dtau[iv], 
			  wc->optprop.g1[iv], wc->optprop.ssa[iv], 
			  output->wc.zd, &input.wc.layer);
      
      break;
    }
    
    break;
    
  case WCPROP_MIE:
  case WCPROP_FILE:

    /* Read optical properties file */
    status = read_wcloud_prop (input, output);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading water cloud property file %s\n", 
	       status, input.filename[FN_WCLOUD_PROP]);
      return status;
    }

    break;

  default:
    fprintf (stderr, "Error, unknown water cloud property %d. This is not\n", input.wc.properties);
    fprintf (stderr, "supposed to happen and indicates a coding error.\n");

    return -1;
    break;
  }

  return 0;
}
  




/* Read ice cloud data from file using read_cld_file() and convert */
/* microphysical properties to optical properties                  */

int read_and_convert_ic_file (input_struct input, output_struct *output,
			      char *filename, cld_out_struct *ic)
{
  int iv=0, status=0;

  status = read_cld_file (input, output->wl.nlambda_t, filename, ic, 1);
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  
  switch (input.ic.properties) {
  case ICPROP_YANG:
    if (!input.quiet)
      fprintf (stderr, " ... using Yang et al. [2000] ice cloud properties\n");
    
    for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
      
      status = ic_yang (output->wl.lambda_t[iv], ic->nlev-1,
			input.ic.habit,
			input.filename[FN_PATH],
			ic->microphys.lwc, ic->microphys.effr, 
			ic->optprop.dtau[iv], 
			ic->optprop.ff[iv], ic->optprop.g1[iv], ic->optprop.g2[iv], 
			ic->optprop.ssa[iv], 
			output->ic.zd, input.ic.layer);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ic_yang()\n", status);
	return status;
      }
    }
    break;
    
  case ICPROP_FU:
    
    if (!input.quiet)
      fprintf (stderr, " ... using Fu et al. [1998] ice cloud properties\n");
    
    /* calculate cloud optical properties */
    switch (input.ck_scheme) {
    case CK_FU:
      
#if HAVE_FULIOU
      status = ckdfuice (ic->zd, ic->microphys.effr, ic->microphys.lwc, ic->nlev,
			 ic->optprop.dtau, ic->optprop.g1, ic->optprop.ssa);
#else
      fprintf (stderr, "Error, Fu and Liou not supported!\n");
      return -1;
#endif
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by ckdfuice()\n", status);
	return status;
      }
      
      break;
      
    default:
      
      for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
	
	if (output->wl.lambda_t[iv]<=4000.0) {
	  
	  status = ic_fu96 (output->wl.lambda_t[iv], ic->nlev-1,
			    input.filename[FN_PATH],
			    ic->microphys.lwc, ic->microphys.effr, 
			    ic->optprop.dtau[iv], ic->optprop.g1[iv], ic->optprop.ssa[iv], ic->optprop.fdel[iv],
			    output->ic.zd, input.ic.layer);

	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by ic_fu96()\n", status);
	    return status;
	  }
	}
	else {
	  status = ic_fu98 (output->wl.lambda_t[iv], ic->nlev-1,
			    input.filename[FN_PATH],
			    ic->microphys.lwc, ic->microphys.effr, 
			    ic->optprop.dtau[iv], ic->optprop.g1[iv], ic->optprop.ssa[iv], 
			    output->ic.zd, input.ic.layer);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by ic_fu98()\n", status);
	    return status;
	  }
	}
      }
    }
    break;

  case ICPROP_FILE:
  case ICPROP_MIE:
    
    /* Read optical properties file */
    status = read_icloud_prop (input, output);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading ice cloud property file %s\n", 
	       status, input.filename[FN_ICLOUD_PROP]);
      return status;
    }

    break;

  default:
    fprintf (stderr, "Error, unknown ice cloud property %d. This is not\n", input.ic.properties);
    fprintf (stderr, "supposed to happen and indicates a coding error.\n");

    return -1;
    break;
  }

  return 0;
}



/* convert microphysical and single scattering properties */
/* to optical properties at model layers                  */

int ssprop2optprop (cld_out_struct *cld, int iv)
{

  int lc=0, status=0;
  
  
  for (lc=0; lc<cld->nlev-1; lc++) {  /* count layers, 0 ... nlev-2 */
    
    /* first free memory, if already allocated */
    if (cld->optprop.moment[iv][lc] != NULL) {
      free (cld->optprop.moment[iv][lc]);
      cld->optprop.moment[iv][lc] = NULL;
    }

    if (cld->microphys.lwc[lc+1] > 0) {
      
      /* interpolate Legendre coefficients to specified effective radius; */
      /* memory for moments is allocated here in this function.           */
      status = interpolate_legendre (cld->ssprop.legen[iv], 
				     cld->ssprop.nleg[iv], 
				     cld->ssprop.nreff[iv],
				     cld->ssprop.r0[iv], 
				     cld->ssprop.dr[iv],
				     cld->ssprop.extinc[iv], 
				     cld->ssprop.albedo[iv],
				     cld->microphys.effr[lc+1],
				     &(cld->optprop.moment[iv][lc]), 
				     &(cld->optprop.nmom[iv][lc]), 
				     &(cld->optprop.dtau[iv][lc]), 
				     &(cld->optprop.ssa[iv][lc]));
      
      if (status!=0) {
	fprintf (stderr, "Error %d interpolating Legendre coefficients to reff = %f in layer %d.\n", 
		 status, cld->microphys.effr[lc+1], lc);
	return status;
      }
	  

      /* convert from extinction / lwc to optical thickness dtau */
      cld->optprop.dtau[iv][lc] *= cld->microphys.lwc[lc+1] * 
	(cld->zd[lc] - cld->zd[lc+1]);
      
      if (cld->optprop.nmom[iv][lc]>1)
	cld->optprop.g1[iv][lc] = cld->optprop.moment[iv][lc][1];

    }
    else {
      cld->optprop.dtau[iv][lc] = 0;
      cld->optprop.ssa [iv][lc] = 0;
      cld->optprop.g1  [iv][lc] = 0;

      /* reset Legendre coefficients */
      cld->optprop.nmom[iv][lc] = 0;
    }
  }
  

  return 0;
}




int read_wcloud_prop (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0;

  char momfilename[FILENAME_MAX]="";
  char tempstr[20]="";

  double wavelen=0, nre=0, nim=0;
  
  if (input.verbose) {
    fprintf (stderr, " ... reading water cloud optical properties from %s\n", 
	     input.filename[FN_WCLOUD_PROP]);
  }

  for (iv=0; iv<output->wl.nlambda_t; iv++) {
    
    /* read optical properties from file */
    strcpy (momfilename, input.filename[FN_WCLOUD_PROP]);
    
    if (iv+1<100)
      strcat (momfilename, "0");
    
    if (iv+1<10) 
      strcat (momfilename, "0");
    
    sprintf (tempstr, "%d", iv+1);
    
    strcat (momfilename, tempstr);
    strcat (momfilename, ".mie");
	
    status = read_mie_table (momfilename, 
			     &(output->wc.ssprop.r0[iv]), 
			     &(output->wc.ssprop.dr[iv]), 
			     &(output->wc.ssprop.nreff[iv]),
			     &wavelen, &nre, &nim, 
			     &(output->wc.ssprop.extinc[iv]), 
			     &(output->wc.ssprop.albedo[iv]), 
			     &(output->wc.ssprop.nleg[iv]),
			     &(output->wc.ssprop.legen[iv]), 
			     input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, momfilename);
      return status;
    }
	
    output->wc.ssprop.alloc = 1;
    
    
    /* convert microphysical and single scattering properties */
    /* to optical properties                                  */
    status = ssprop2optprop (&(output->wc), iv);
    
    if (status!=0) {
      fprintf (stderr, "Error %d calculating water cloud optical properties\n", status);
      return status;
    }
  }	
  
  /* increase number of moments if more found */
  for (iv=0; iv<output->wl.nlambda_t; iv++)
    for (lc=0; lc<output->wc.nlev-1; lc++)
      if (output->wc.optprop.nmom[iv][lc]-1 > output->atm.nmom)
	output->atm.nmom = output->wc.optprop.nmom[iv][lc]-1;


  return 0;
}
  






int read_icloud_prop (input_struct input, output_struct *output)
{
  int status=0, lc=0, iv=0;

  char momfilename[FILENAME_MAX]="";
  char tempstr[20]="";

  double wavelen=0, nre=0, nim=0;
  
      

  if (input.verbose) {
    fprintf (stderr, " ... reading ice cloud optical properties from %s\n", 
	     input.filename[FN_ICLOUD_PROP]);
  }

  for (iv=0; iv<output->wl.nlambda_t; iv++) {
    
    /* read optical properties from file */
    strcpy (momfilename, input.filename[FN_ICLOUD_PROP]);
    
    if (iv+1<100)
      strcat (momfilename, "0");
    
    if (iv+1<10) 
      strcat (momfilename, "0");
    
    sprintf (tempstr, "%d", iv+1);
    
    strcat (momfilename, tempstr);
    strcat (momfilename, ".mie");
    
    status = read_mie_table (momfilename, 
			     &(output->ic.ssprop.r0[iv]), 
			     &(output->ic.ssprop.dr[iv]), 
			     &(output->ic.ssprop.nreff[iv]),
			     &wavelen, &nre, &nim, 
			     &(output->ic.ssprop.extinc[iv]), 
			     &(output->ic.ssprop.albedo[iv]), 
			     &(output->ic.ssprop.nleg[iv]),
			     &(output->ic.ssprop.legen[iv]), 
			     input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, momfilename);
      return status;
    }
    
    output->ic.ssprop.alloc = 1;
    
    /* convert microphysical and single scattering properties */
    /* to optical properties                                  */
    status = ssprop2optprop (&(output->ic), iv);
    
    if (status!=0) {
      fprintf (stderr, "Error %d calculating ice cloud optical properties\n", status);
      return status;
    }
  }	
  
  /* increase number of moments if more found */
  for (iv=0; iv<output->wl.nlambda_t; iv++)
    for (lc=0; lc<output->ic.nlev-1; lc++)
      if (output->ic.optprop.nmom[iv][lc]-1 > output->atm.nmom )
	output->atm.nmom = output->ic.optprop.nmom[iv][lc]-1;
  
  return 0;
}





/* overwrite water cloud properties with user-defined properties */
int apply_user_defined_properties_to_cld (cld_inp_struct cldin, 
					  atm_out_struct atmout,
					  wl_out_struct  wlout,
					  float altitude,
					  cld_out_struct *cldout)
{
  int iv=0, lc=0, status=0;
  float scale_factor=0, sum=0;
  double *tautot = NULL, *lambda = NULL;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double tau550=0;



  for (iv=0; iv<wlout.nlambda_t; iv++) {
    
    /* Set the cloud optical thickness to a constant user-defined value */
    if (cldin.tau >= 0.0) {
      sum = 0.0;
      for (lc=0; lc<cldout->nlev-1; lc++) {
	if (cldout->zd[lc] > altitude) {    /* layers with upper boundary above the surface */
	  if (cldout->zd[lc+1] > altitude)  /* layers completely above the surface          */
	    sum += cldout->optprop.dtau[iv][lc];
	  else                              /* upper boundary above, lower boundary below surface */
	    sum += cldout->optprop.dtau[iv][lc] * 
	      (cldout->zd[lc]-altitude) / (cldout->zd[lc]-cldout->zd[lc+1]);
	}
      }

      scale_factor = 0.0;
      if (sum>0)
	scale_factor = cldin.tau/sum;
      
      for (lc=0; lc<cldout->nlev-1; lc++)
	cldout->optprop.dtau[iv][lc] *= scale_factor;
    }
    
    /* Set the cloud asymmetry factor to a user-defined value */
    if (cldin.gg >= -1.0)
      for (lc=0; lc<cldout->nlev-1; lc++)
	cldout->optprop.g1[iv][lc] = cldin.gg;

    /* Set the cloud single scattering albedo to a user-defined value */
    if (cldin.ssa >= 0.0)
      for (lc=0; lc<cldout->nlev-1; lc++)
	cldout->optprop.ssa[iv][lc] = cldin.ssa;

    /* Scale the cloud asymmetry factor with a user-defined value */
    if (cldin.scale_gg >= 0.0)
      for (lc=0; lc<cldout->nlev-1; lc++) {
	cldout->optprop.g1[iv][lc] *= cldin.scale_gg;
	if (cldout->optprop.g1[iv][lc]>1.0)
	  cldout->optprop.g1[iv][lc] = 1.0;
      }

    /* Scale the cloud single scattering albedo with a user-defined value */
    if (cldin.scale_ssa >= 0.0)
      for (lc=0; lc<cldout->nlev-1; lc++) {
	cldout->optprop.ssa[iv][lc] *= cldin.scale_ssa;
	if (cldout->optprop.ssa[iv][lc]>1.0)
	  cldout->optprop.ssa[iv][lc] = 1.0;
      }
  }



  /* Set the cloud optical thickness to a user-defined value at 550nm */
  if (cldin.tau550 >= 0.0) {
    lambda = calloc (wlout.nlambda_t, sizeof(double));
    tautot = calloc (wlout.nlambda_t, sizeof(double));

    for (iv=0; iv<wlout.nlambda_t; iv++) {
      lambda[iv] = (double) wlout.lambda_t[iv];

      for (lc=0; lc<cldout->nlev-1; lc++) {
	if (cldout->zd[lc] > altitude) {    /* layers with upper boundary above the surface */
	  if (cldout->zd[lc+1] > altitude)  /* layers completely above the surface          */
	    tautot[iv] += cldout->optprop.dtau[iv][lc];
	  else                              /* upper boundary above, lower boundary below surface */
	    tautot[iv] += cldout->optprop.dtau[iv][lc] * 
	      (cldout->zd[lc]-altitude) / (cldout->zd[lc]-cldout->zd[lc+1]);
	}
      }
    }

    status = linear_coeffc (lambda, tautot, wlout.nlambda_t, &a0, &a1, &a2, &a3);
    if (status!=0) {
      fprintf (stderr, "Error %d calculating interpolation coefficients for cloud optical thickness.\n", 
	       status);
      return status;
    }

    status = calc_splined_value (550.0, &tau550, lambda, wlout.nlambda_t, a0, a1, a2, a3);
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating cloud optical thickness to 550nm.\n", 
	       status);
      fprintf (stderr, "In the current implementation of xx_set_tau550, it is required\n");
      fprintf (stderr, "that the internal wavelength range includes 550nm.\n");
      fprintf (stderr, "Hope to change that in future!\n");
      return status;
    }
    
    if (tau550>0.0) {
      scale_factor = cldin.tau550 / tau550;
      
      for (iv=0; iv<wlout.nlambda_t; iv++)
	for (lc=0; lc<cldout->nlev-1; lc++)
	  cldout->optprop.dtau[iv][lc] *= scale_factor;
    }


    free(a0); free(a1); free(a2); free(a3);
    free(lambda); free(tautot);
  }

  return 0;
}


/*agf*/
static int ic_yang (float wavelength, int nlyr, int habit,
		    char *path,
		    float *iwc, float *reff, 
		    float *dtau, 
		    float *f, float *g1, float *g2, float *ssa, 
		    float *zd, int layer)
{
  float ext=0;
  int lc=0, status=0;
  char mipath[256]="";
  float maxreff = 55.0;
  
  sprintf(mipath,"%s/ic/yang56/",path);

  if (layer==0) {
    fprintf (stderr, "Error, Fu et al. [1996] can only be used with cloud properties\n");
    fprintf (stderr, "defined per layer.\n");
    return -1;
  }

  for (lc=0; lc<nlyr; lc++) {
  
    /* calculate dtau[lc], g1[lc], g2[lc], f[lc], ssa[lc] from */
    /* iwc[lc], reff[lc], wavelength                           */

    if (iwc[lc+1]>0.0) {

#if HAVE_YANG56
      switch (habit)  {
      case IC_HABIT_SOLIDCOLUMN:   
	maxreff=MAX_REFF_SC; 
	break;
      case IC_HABIT_HOLLOWCOLUMN:  
	maxreff=MAX_REFF_HC; 
	break;
      case IC_HABIT_AGGREGATE:     
	maxreff=MAX_REFF_RA; 
	break;
      case IC_HABIT_ROSETTE4:      
	maxreff=MAX_REFF_R4; 
	break;
      case IC_HABIT_ROSETTE6:      
	maxreff=MAX_REFF_R6; 
	break;
      case IC_HABIT_PLATE:         
	maxreff=MAX_REFF_PL; 
	break;
      default:
	fprintf (stderr, "Error, unknown habit %d\n", habit);
	return -1;
      }
      
      if (reff[lc+1]>maxreff) {
	fprintf (stderr, "Error, ice effective radius too large for this habit\n");
	return -1;
      }
      

      status = yang56 (wavelength/1000.0, reff[lc+1], habit, mipath, &ext, ssa+lc, g1+lc, g2+lc, f+lc);
      
      if (status!=0) {
	fprintf (stderr, "Error %d returned by yang56()\n", status);
	return status;
      }
#else
      fprintf (stderr, " ***********************************************************************\n");
      fprintf (stderr, " * You have built uvspec without Yang et al support and hence cannot   *\n");
      fprintf (stderr, " * use 'ic_properties yang'.                                           *\n");
      fprintf (stderr, " ***********************************************************************\n");
      return -1;
#endif

      dtau[lc] = ext*iwc[lc+1]*(zd[lc]-zd[lc+1]);
    }
  }
  
  return 0;   /* if ok */
}




/*******************************************************************/
/* Ice cloud properties according to                               */
/* Q. Fu, An Accurate Parameterization of the Solar Radiative      */
/* Properties of Cirrus Clouds for Climate Models,                 */
/* Journal of Climate 9, 2058-2082, 1996.                          */
/*******************************************************************/

int ic_fu96 (float wavelength, int nlyr,
	     char *path,
	     float *iwc, float *reff, 
	     float *dtau, float *g, float *ssa, float *fdel,
	     float *zd, int layer)
{
  int i=0;
  int lc=0, status=0;
  char filename [FILENAME_MAX]="";

  static double *al1=NULL, *al2=NULL, *a0=NULL, *a1=NULL;
  static double *bl1=NULL, *bl2=NULL, *b0=NULL, *b1=NULL, *b2=NULL, *b3=NULL;
  static double *cl1=NULL, *cl2=NULL, *c0=NULL, *c1=NULL, *c2=NULL, *c3=NULL;
  static double *dl1=NULL, *dl2=NULL, *d0=NULL, *d1=NULL, *d2=NULL, *d3=NULL;

  static int na=0, nb=0, nc=0, nd=0;

  static int first=1;

  double D=0;
  double asy=0, ext=0, coalb=0, del=0;
  double f=0, fw=0;
  double taunp=0, ssanp=0, gnp=0;

  float *deff = calloc (nlyr+1, sizeof(float));

  if (layer==0) {
    fprintf (stderr, "Error, Fu et al. [1996] can only be used with cloud properties\n");
    fprintf (stderr, "defined per layer.\n");
    return -1;
  }

  if (first) {  /* read only once */
    first=0;
    
    /* read extinction data */
    sprintf (filename,"%s/ic/fu96/fu96.ext",path);
    status = read_4c_file (filename, &al1, &al2, &a0, &a1, &na);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu [1996] data from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read absorption data */
    sprintf (filename,"%s/ic/fu96/fu96.ssa",path);
    status = read_6c_file (filename, &bl1, &bl2, &b0, &b1, &b2, &b3, &nb);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu [1996] data from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read asymmetry parameter data */
    sprintf (filename,"%s/ic/fu96/fu96.asy",path);
    status = read_6c_file (filename, &cl1, &cl2, &c0, &c1, &c2, &c3, &nc);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu [1996] data from %s\n", 
	       status, filename);
      return status;
    }

    /* read delta-fraction parameter data */
    sprintf (filename,"%s/ic/fu96/fu96.del",path);
    status = read_6c_file (filename, &dl1, &dl2, &d0, &d1, &d2, &d3, &nd);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu [1996] data from %s\n", 
	       status, filename);
      return status;
    }

    /* check if wavelength grids are identical */
    if (na!=nb || na!=nc || na!=nd) {
      fprintf (stderr, "Error, something weird happened to the Fu et al. [1998] data files.\n");
      fprintf (stderr, "Found different wavelength grids.\n");
      return -1;
    }

    for (i=0; i<na; i++)
      if (al1[i] != bl1[i] || al1[i] != cl1[i] || al1[i] != dl1[i]) {
	fprintf (stderr, "Error, something weird happened to the Fu et al. [1998] data files.\n");
	fprintf (stderr, "Found different wavelength grids.\n");
	return -1;
      }
    
    /* convert from micron to nm */
    for (i=0; i<na; i++) {
      al1[i] *= 1000.0;
      al2[i] *= 1000.0;
    }
  }
  
  /* determine the tabulated wavelength interval containing the actual wavelength */
  if (wavelength < al1[0] || wavelength > al2[na-1]) {
    fprintf (stderr, "Error, wavelength %f nm not covered by Fu et al. [1998]\n", wavelength);
    fprintf (stderr, "Allowed range is %.1f - %.1f nm\n", al1[0], al2[na-1]);
    return -1;
  }


  /* Fu et al. requires effective diameter */
  for (lc=0; lc<=nlyr; lc++)
    deff[lc] = 2.0 * reff[lc];


  for (i=0; i<na-1; i++)
    if (wavelength < al2[i])
      break;

  for (lc=0; lc<nlyr; lc++) {
    if (iwc[lc+1]>0) {
      /* calculate dtau[lc], g[lc], and ssa[lc] from */
      /* iwc[lc], deff[lc], wavelength               */
      
      D = deff[lc+1];
      
      /* extinction */
      ext = a0[i] + a1[i]/D;
      
      /* single scattering albedo */
      coalb = b0[i] + b1[i] * D + b2[i] * D * D + b3[i] * D * D * D;
      
      /* asymmetry parameter */
      asy = c0[i] + c1[i] * D + c2[i] * D * D + c3[i] * D * D * D;

      /* delta-fraction */
      del = d0[i] + d1[i] * D + d2[i] * D * D + d3[i] * D * D * D;
      
      dtau[lc] = ext*iwc[lc+1]*(zd[lc]-zd[lc+1])*1000.0;
      ssa [lc] = 1.0-coalb;
      g   [lc] = asy;
      fdel[lc] = del;

      taunp = dtau[lc];
      ssanp = ssa [lc];
      gnp   = g   [lc];

      f  = 0.5 / ssa[lc] + fdel[lc];
      fw = f * ssa[lc];

      dtau[lc] = (1.0 - fw) * dtau[lc];
      ssa [lc] = (1.0 - f) * ssa[lc] / (1.0 - fw);
      g   [lc] = (g[lc] - f) / (1.0 - f);


    }
  }

    free(deff);

  return 0;   /* if ok */
}

/*******************************************************************/
/* Ice cloud properties according to                               */
/* Q. Fu, P. Yang, and W.B. Sun, An Accurate Parameterization of   */
/* the Infrared Radiative Properties of Cirrus Clouds for Climate  */
/* Models, Journal of Climate 11, 2223-2237, 1998.                 */
/*******************************************************************/

int ic_fu98 (float wavelength, int nlyr,
	     char *path,
	     float *iwc, float *reff, 
	     float *dtau, float *g, float *ssa, 
	     float *zd, int layer)
{
  int i=0, lower=0, upper=0;
  int lc=0, status=0;
  char filename [FILENAME_MAX]="";

  static double *al=NULL, *a0=NULL, *a1=NULL, *a2=NULL;
  static double *bl=NULL, *b0=NULL, *b1=NULL, *b2=NULL, *b3=NULL;
  static double *cl=NULL, *c0=NULL, *c1=NULL, *c2=NULL, *c3=NULL;

  static int na=0, nb=0, nc=0;

  static int first=1;

  double D=0;
  double extlower=0, extupper=0, ext=0;
  double abslower=0, absupper=0, abs=0;
  double asylower=0, asyupper=0, asy=0;

  float *deff = calloc (nlyr+1, sizeof(float));

  if (layer==0) {
    fprintf (stderr, "Error, Fu et al. [1998] can only be used with cloud properties\n");
    fprintf (stderr, "defined per layer.\n");
    return -1;
  }

  if (first) {  /* read only once */
    first=0;
    
    /* read extinction data */
    sprintf (filename,"%s/ic/fu98/fu98.ext",path);
    status = read_4c_file (filename, &al, &a0, &a1, &a2, &na);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu et al. [1998] data from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read absorption data */
    sprintf (filename,"%s/ic/fu98/fu98.abs",path);
    status = read_5c_file (filename, &bl, &b0, &b1, &b2, &b3, &nb);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu et al. [1998] data from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read asymmetry parameter data */
    sprintf (filename,"%s/ic/fu98/fu98.asy",path);
    status = read_5c_file (filename, &cl, &c0, &c1, &c2, &c3, &nc);
    if (status!=0) {
      fprintf (stderr, "Error %d reading Fu et al. [1998] data from %s\n", 
	       status, filename);
      return status;
    }

    /* check if wavelength grids are identical */
    if (na!=nb || na!=nc) {
      fprintf (stderr, "Error, something weird happened to the Fu et al. [1998] data files.\n");
      fprintf (stderr, "Found different wavelength grids.\n");
      return -1;
    }

    for (i=0; i<na; i++)
      if (al[i] != bl[i] || al[i] != cl[i]) {
	fprintf (stderr, "Error, something weird happened to the Fu et al. [1998] data files.\n");
	fprintf (stderr, "Found different wavelength grids.\n");
	return -1;
      }
    
    /* convert from micron to nm */
    for (i=0; i<na; i++)
      al[i] *= 1000.0;
    
  }

  
  /* determine tabulated wavelengths above and below the actual wavelength */
  if (wavelength < al[0] || wavelength > al[na-1]) {
    fprintf (stderr, "Error, wavelength %f nm not covered by Fu et al. [1998]\n", wavelength);
    fprintf (stderr, "Allowed range is %.1f - %.1f nm\n", al[0], al[na-1]);
    return -1;
  }


  /* Fu et al. requires effective diameter */
  for (lc=0; lc<=nlyr; lc++)
    deff[lc] = 2.0 * reff[lc];


  for (i=0; i<na-1; i++)
    if (wavelength < al[i])
      break;

  lower=i-1;
  upper=i;
    
  for (lc=0; lc<nlyr; lc++) {
  

    if (iwc[lc+1]>0) {
      /* calculate dtau[lc], g[lc], and ssa[lc] from */
      /* iwc[lc], deff[lc], wavelength               */
      
      D = deff[lc+1];
      
      /* extinction */
      extlower = a0[lower] + a1[lower]/D + a2[lower] / (D * D);
      extupper = a0[upper] + a1[upper]/D + a2[upper] / (D * D);
      ext = extlower + (wavelength - al[lower]) / (al[upper]-al[lower]) * (extupper-extlower);
      
      /* absorption */
      abslower = b0[lower] + b1[lower] * D + b2[lower] * D * D + b3[lower] * D * D * D;
      absupper = b0[upper] + b1[upper] * D + b2[upper] * D * D + b3[upper] * D * D * D;
      abs = (abslower + (wavelength - al[lower]) / (al[upper]-al[lower]) * (absupper-abslower)) / D;
      
      /* asymmetry parameter */
      asylower = c0[lower] + c1[lower] * D + c2[lower] * D * D + c3[lower] * D * D * D;
      asyupper = c0[upper] + c1[upper] * D + c2[upper] * D * D + c3[upper] * D * D * D;
      asy = asylower + (wavelength - al[lower]) / (al[upper]-al[lower]) * (asyupper-asylower);
      
      dtau[lc] = ext*iwc[lc+1]*(zd[lc]-zd[lc+1])*1000.0;
      ssa [lc] = (ext-abs)/ext;
      g   [lc] = asy;
    }
  }

  free(deff);
  
  return 0;   /* if ok */
}




