/*--------------------------------------------------------------------
 * $Id: extraterrestrial.c,v 1.14 2002/12/09 23:29:19 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>

#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/sun.h"

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif


/************************************/
/* prototypes of internal functions */
/************************************/

static int read_extraterrestrial(char* extraterrestrial_filename, 
				 wl_inp_struct *wl_inp, wl_out_struct *wl_out,
				 float wl_start, float wl_end, int cdf, int ck_scheme);




/**************************************************************/
/* Setup extraterrestrial irradiance.                         */
/**************************************************************/

int setup_extraterrestrial (input_struct input, output_struct *output)
{
  int status=0, iv=0, cdf=0;
  char temp[FILENAME_MAX]="";
  char filename[FILENAME_MAX]="";

  float *x_flt=NULL, *y_flt=NULL;
  int n_flt=0;

  /* For special applications, pre-defined extraterrestrial */
  /* spectra are used.                                      */

  /* construct location of the extraterrestrial flux */
  strcpy (temp, input.filename[FN_PATH]);
  switch (input.ck_scheme) {
  case CK_KATO:
  case CK_KATO2:
    strcat (temp, "solar_flux/kato");
    break;

  case CK_FU:
    strcat (temp, "solar_flux/fu");
    break;

  case CK_AVHRR_KRATZ:
    strcat (temp, "solar_flux/kratz");
    break;

  default:
    break;
  }
    

  switch (input.ck_scheme) {
  case CK_KATO:
  case CK_KATO2:
  case CK_FU:
  case CK_AVHRR_KRATZ:
    
    /* check if we need to replace the extraterrestrial flux */
    if (!strcmp(temp, input.filename[FN_EXTRATERRESTRIAL]))
      strcpy (filename, input.filename[FN_EXTRATERRESTRIAL]);
    else 
      strcpy (filename, temp);
    
    if (!input.quiet)
      fprintf (stderr, " ... replaced extraterrestrial flux by %s\n", filename);
    
    cdf=0;

    break;

  case CK_GENERIC:
    if (!input.quiet)
      if (strlen(input.filename[FN_EXTRATERRESTRIAL]) > 0)
	fprintf (stderr, " ... ignoring %s, reading extraterrestrial flux from %s\n",
		 input.filename[FN_EXTRATERRESTRIAL], input.filename[FN_CK_GENERIC]);


    strcpy (filename, input.filename[FN_CK_GENERIC]);
    cdf=1;
    break;

  case CK_NONE:
  case CK_SBDART:
    strcpy (filename, input.filename[FN_EXTRATERRESTRIAL]);
    break;

  default:
    fprintf (stderr, "Error: unsupported correlated-k scheme\n");
    return -1;
  }


  
  /**** Read extraterrestrial irradiance ****/
  status = read_extraterrestrial(filename, 
				 &input.wl, &output->wl, 
				 output->wl.start, output->wl.end, cdf, input.ck_scheme);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }
  

  /* allocate memory for filter function in any case */
  output->wl.filter = calloc (output->wl.nlambda_h, sizeof(float));
  
  /* set filter function to 1 (default) */
  for (iv=0; iv<output->wl.nlambda_h; iv++)
    output->wl.filter[iv] = 1;


  /**** Read filter function ****/
  if (strlen(input.filename[FN_FILTERFUNCTION])>0) {

    fprintf (stderr, " ... reading filter function from  %s\n",
	     input.filename[FN_FILTERFUNCTION]);

    status = read_2c_file_float (input.filename[FN_FILTERFUNCTION], &x_flt, &y_flt, &n_flt); 
    if (status!=0) {
      fprintf (stderr, "Error %d reading filter function from %s\n", 
	       status, input.filename[FN_FILTERFUNCTION]);
      return status;
    }

    status = arb_wvn (n_flt, x_flt, y_flt,  
		      output->wl.nlambda_h, output->wl.lambda_h, output->wl.filter, 1, 0);

    if (status!=0) {
      fprintf (stderr, "Error %d interpolating filter function\n", status);
      return status;
    }

    /* multiply extraterrestrial flux with filter function */
    // for (iv=0; iv<output->wl.nlambda_h; iv++)
    //   output->wl.fbeam[iv] *= output->wl.filter[iv];
  }


  /**** Correction factor for Earth-Sun distance ****/
  output->eccentricity_factor = 1.0;
  
  if (input.day_of_year > 0.0) {
    output->eccentricity_factor = eccentricity (input.day_of_year);
    for (iv=0; iv<output->wl.nlambda_h; iv++)
      output->wl.fbeam[iv] *= output->eccentricity_factor;
  }

  return 0;
}





/**************************************************************/
/* Read extraterrestrial irradiance.                          */
/**************************************************************/

static int read_extraterrestrial (char *filename, 
				  wl_inp_struct *wl_inp, wl_out_struct *wl_out,
				  float wl_start, float wl_end, int cdf, int ck_scheme) 
{
  int iv=0, nlambda=0, status=0;
  float *tmp_lambda=NULL, *tmp_fbeam=NULL;

#if HAVE_LIBNETCDF
  double *wvnlo=NULL, *wvnhi=NULL;
  int id_wvnlo=0, id_wvnhi=0;
  int ncid=0;
  int idd_nwvn=0;
  int id_extra=0;
  
  int start[1] = {0};
  int count[1] = {0};
#endif


  /* read extraterrestrial irradiance */
  if (!cdf) {
    status = read_2c_file_float (filename, 
				 &tmp_lambda, &tmp_fbeam, &(wl_out->nlambda_h));
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
  }
  else {

#if HAVE_LIBNETCDF

    /* open netcdf file */
    status = nc_open (filename, NC_NOWRITE, &ncid);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
      return status;
    }

    /* get dimension id for "nwvn" */
    status = nc_inq_dimid (ncid, "nwvn", &idd_nwvn);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
  
    /* get dimension length for "nwvn" */
    status = nc_inq_dimlen (ncid, idd_nwvn, &(wl_out->nlambda_h));
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    start[0] = 0;
    count[0] = wl_out->nlambda_h;
    
    /* allocate memory for fields */
    tmp_fbeam  = calloc (wl_out->nlambda_h, sizeof(float));
    tmp_lambda = calloc (wl_out->nlambda_h, sizeof(float));
    wvnlo      = calloc (wl_out->nlambda_h, sizeof(double));
    wvnhi      = calloc (wl_out->nlambda_h, sizeof(double));

    /* get variable id for "wvnlo" */
    status = nc_inq_varid (ncid, "wvnlo", &id_wvnlo);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* read "wvnlo" */
    status = nc_get_vara_double (ncid, id_wvnlo, start, count, wvnlo);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    /* get variable id for "wvnhi" */
    status = nc_inq_varid (ncid, "wvnhi", &id_wvnhi);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* read "wvnhi" */
    status = nc_get_vara_double (ncid, id_wvnhi, start, count, wvnhi);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    for (iv=0; iv<wl_out->nlambda_h; iv++)
      tmp_lambda[iv] = 1.0E7 / (0.5 * (wvnlo[iv] + wvnhi[iv]));



    /* get variable id for "extra" */
    status = nc_inq_varid (ncid, "extra", &id_extra);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading extraterrestrial irradiance from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read "extra" */
    status = nc_get_vara_float (ncid, id_extra, start, count, tmp_fbeam);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    nc_close(ncid);

#else
  fprintf (stderr, " ***********************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot            *\n");
  fprintf (stderr, " * use the generic correlated_k option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ***********************************************************************\n");
  return -1;
#endif

  }

  /* count the extraterrestrial wavelengths within the */
  /* user-defined wavelength range                     */
  nlambda = 0;
  for (iv=0; iv<wl_out->nlambda_h; iv++)
    if ((tmp_lambda[iv]>=wl_start) && (tmp_lambda[iv]<=wl_end))
      nlambda++;

  wl_out->lambda_h = (float *) calloc (nlambda, sizeof(float));
  wl_out->fbeam    = (float *) calloc (nlambda, sizeof(float));

  wl_out->map_e2h  = (int *)   calloc (nlambda, sizeof(int));

  if ( ck_scheme == CK_NONE ) {
    /* Check that the extraterrestrial spectrum did cover the whole wanted */
    /* wavelength range                                                    */
    if ( wl_start < tmp_lambda[0] ) {
      fprintf (stderr, "Error: wanted start wavelength: %f outside specified extraterrestrial\n",wl_start);
      fprintf (stderr, "       spectrum range [%f-%f]\n", tmp_lambda[0], tmp_lambda[wl_out->nlambda_h-1]);
      return -1;
    }
    if ( wl_end > tmp_lambda[wl_out->nlambda_h-1] ) {
      fprintf (stderr, "Error: wanted end wavelength: %f outside specified extraterrestrial\n", wl_end);
      fprintf (stderr, "       spectrum range [%f-%f]\n", tmp_lambda[0], tmp_lambda[wl_out->nlambda_h-1]);
      return -1;
    }
  }

  nlambda = 0;
  for (iv=0; iv<wl_out->nlambda_h; iv++)
    if ((tmp_lambda[iv]>=wl_start) && (tmp_lambda[iv]<=wl_end)) {
      wl_out->lambda_h[nlambda] = tmp_lambda[iv];
      wl_out->fbeam[nlambda]    = tmp_fbeam[iv];
      wl_out->map_e2h[nlambda]  = iv;
      
      nlambda++;
    }

  wl_out->nlambda_h = nlambda;

  /* free memory */
  free(tmp_lambda);
  free(tmp_fbeam);

  return 0;
}



